require(rpart)
require(randomForest)
require(MultivariateRandomForest)
require(boot)
require(plyr)
require(hydroGOF)
require(MASS)


###SIMULATION STUDY 4###
##MULTIVARIATE MODEL WITH NON-LINEAR DEPENDENCIES AND ERROR CORRELATION##
###NO. OF COVARIATES = 15###
#USER-DEFINED INPUTS
#specify where you want results to store
filepath = './'

##CREATE PANEL OF 300 DATA POINTS##
n = 100  #CHANGE THIS TO n = 100 FOR CODE TESTING. The training set in Sikdar et al. has 300 data points

#CREATE THE NUMBER OF TREES
ntree = 10 #CHANGE THIS TO ntree = 10 FOR CODE TESTING. In the simulation results in Sikdar et al. this is set at 3000

#CREATE THE NUMBER OF FORESTS
nforest = 1 #CHANGE THIS TO nforest = 1 FOR CODE TESTING. In the simulation results in Sikdar et al. this is set at 10


##CREATING CONTINUOUS COVARIATES####
##ASSUMING CONT. COVARIATES GENERATED FROM UNIFORM DISTRIBUTION##
x1 = runif(n,0,1)
x3 = rpois(n,50)
x5 = runif(n,0,1)
x7 = runif(n,0,1)
x9 = runif(n,0,0.5)
x11 = runif(n,1,1)
x13 = runif(n,0,0.25)
x15 = runif(n,0,0.5)

##CREATING 1/0 INDICATOR COVARIATES####
##USING RBINOM ON PROBABILITIES GENERATED FROM UNIFORM DISTRIBUTION##

x2 = rbinom(n,1,0.5)
x4 = rbinom(n,1,runif(n,0,0.5))
x6 = rbinom(n,1,runif(n,0,0.4))
x8 = rbinom(n,1,runif(n,0,0.4))
x10 = rbinom(n,1,runif(n,0,0.3))
x12 = rbinom(n,1,runif(n,0,0.3))
x14 = rbinom(n,1,runif(n,0,0.25))

##CREATING MULTIVARIATE OBJECTIVE FUNCTION ###
##dependencies only on the first five covariates

#coefficients for Y1
a11 = 1.85
a12 = 0.95
a13 = -0.05
a14 = 0.95
a15 = -0.85
#coefficients for Y2
a21 = 1.7
a22 = 0.9
a23 = 0.04
a24 = 0.8
a25 = -0.75
#coefficients for Y3
a31 = 2.45
a32 = 1
a33 = 0.04
a34 = 0.95
a35 = -0.9
#coefficients for Y4
a41 = 2
a42 = 0.9
a43 = -0.04
a44 = 0.8
a45 = 0.75

####DERIVING DATA FOR nforest = 10 FORESTS####

for (num_forest in (1:nforest))

{

##DERIVING explained part of y VALUES

ey1 = a11*x1 + a12*x2 + a13*x3 + a14*x4 + a15*x5 
ey2 = a21*x1 + a22*x2 + a23*x3 + a24*x4 + a25*x5 
ey3 = a31*x1 + a32*x2 + a33*x3 + a34*x4 + a35*x5 
ey4 = a41*x1 + a42*x2 + a43*x3 + a44*x4 + a45*x5 

#INTRODUCING CORRELATIONS IN ERROR TERMS
e4 = rnorm(n,0,var(ey4)/10)
e3 = (rnorm(n,0,var(ey3)/10) + 0.5*e4)*sqrt((var(ey3)/10)/(var(ey3)/10 + 0.5^2*var(ey4)/10))
e2 = (rnorm(n,0,var(ey2)/10) - 0.02*e3)*sqrt((var(ey2)/10)/(var(ey2)/10 + 0.02^2*var(ey3)/10))
e1 = (rnorm(n,0,var(ey1)/10) + 0.2*e2)*sqrt((var(ey1)/10)/(var(ey1)/10 + 0.2^2*var(ey2)/10))

##DERIVING logistic transformation of y VALUES

p1 = plogis(a11*x1 + a12*x2 + a13*x3 + a14*x4 + a15*x5 + e1)
p2 = plogis(a21*x1 + a22*x2 + a23*x3 + a24*x4 + a25*x5 + e2)
p3 = plogis(a31*x1 + a32*x2 + a33*x3 + a34*x4 + a35*x5 + e3)
p4 = plogis(a41*x1 + a42*x2 + a43*x3 + a44*x4 + a45*x5 + e4) 

I1 = rep(0,n)
I2 = rep(0,n)
I3 = rep(0,n)
I4 = rep(0,n)
y1 = rep(0,n)
y2 = rep(0,n)
y3 = rep(0,n)
y4 = rep(0,n)


for (i in (1:n))
{
#DERIVING BINOMIALS (1/0): TO DENOTE WHETHER OR NOT CHANGE HAPPENS
I1[i] = rbinom(1,1,p1[i])
I2[i] = rbinom(1,1,p2[i])
I3[i] = rbinom(1,1,p3[i])
I4[i] = rbinom(1,1,p4[i])

#GENERATING SPARSE DATA FOR Y
y1[i] = I1[i]*rexp(1,1)
y2[i] = I2[i]*rexp(1,1)
y3[i] = I3[i]*rexp(1,1)
y4[i] = I4[i]*rexp(1,1)
}



##full data###

traindata = cbind(y1, y2, y3, y4,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13,x14,x15)


#SPLITTING DATA INTO TRAIN AND TEST SAMPLES

trainY1= traindata[,(1:4)]
trainX1= traindata[,(5:19)]

samplsize = round((2/3)*nrow(trainX1))
dimY1 = ncol(trainY1)
dimX1 = ncol(trainX1)
N = nrow(trainX1)

results1 = matrix(0,samplsize,ntree)  #Initializing matrix of indices of observations used in each tree
cnt_index = matrix(0,N,ntree)         #Initializing matrix of counts of indices of observations used in each tree, i.e., cnt_index[1,1] is the number of times 1st obs was used in tree 1
oob_index = matrix(0,N,ntree)
trainX1_bootstrap = array(c(0,0), dim= c(samplsize,dimX1,ntree))
trainY1_bootstrap = array(c(0,0), dim= c(samplsize,dimY1,ntree))
oob_X1_bootstrap = array(c(0,0), dim= c(N,dimX1, ntree))
oob_Y1_bootstrap = array(c(0,0), dim= c(N, dimY1, ntree))
nobs_oob = rep(0,ntree) #vector for the number of oob samples in ntree trees


for (i in 1:ntree)
{
## A bootstrap sample of indices of observations for each tree
results1[,i]= t(sample(1:N,size = samplsize, replace=TRUE))

for (k in 1:N)
{
cnt_index[k,i] = sum((k == results1[,i])*1)

if (cnt_index[k,i] == 0)
{oob_index[k,i] = k
 oob_X1_bootstrap[k,,i] = trainX1[k,]
 oob_Y1_bootstrap[k,,i] = trainY1[k,]} else {
 oob_index[k,i] = 0
 oob_X1_bootstrap[k,,i] = 0
 oob_Y1_bootstrap[k,,i] = 0
}}

for (j in (1:samplsize))
{ 
##Creating bootstrap samples based on indices of observations used in each tree
trainX1_bootstrap[j,,i] = trainX1[results1[j,i],]
trainY1_bootstrap[j,,i] = trainY1[results1[j,i],]
}

nobs_oob[i] = length(which(rowSums(oob_X1_bootstrap[,,i]) > 0))

}


###Building Trees
###Initializing arrays and matrices
Inv_Cov_array1 = array(c(0,0), dim= c(dimY1,dimY1,ntree))
VIM_matrix1 = matrix(0,dimX1,ntree)
VIM2_matrix = matrix(0,dimX1,ntree) #indicator matrix where (i,j)th value takes 1 if ith covariate is in jth tree, else 0
avg_VIM1 = rep(0,dimX1) #mean variable importance across ntree trees
avg_VIM2 = rep(0,dimX1) #mean variable importance across ntree trees

avg_VIM_SI_test = rep(0,dimX1)    #Mean variable importance computed using the split improvement method
avg_VIM_SI_F_test = rep(0,dimX1)  #Mean variable importance computed using the split improvement method with F statistic hypothesis testing

avg_VIM_SI_train = rep(0,dimX1)    #Mean variable importance computed using the split improvement method
avg_VIM_SI_F_train = rep(0,dimX1)  #Mean variable importance computed using the split improvement method with F statistic hypothesis testing

avg_VIM_Ydiff_test = matrix(0,dimX1,dimY1) #Mean variable importance computed using the response difference method
avg_VIM_Ydiff_F_test = matrix(0,dimX1,dimY1) #Mean variable importance computed using the response difference method with F stat hypothesis testing

avg_VIM_Ydiff_train = matrix(0,dimX1,dimY1) #Mean variable importance computed using the response difference method
avg_VIM_Ydiff_F_train = matrix(0,dimX1,dimY1) #Mean variable importance computed using the response difference method with F stat hypothesis testing

covar_IJ_matrix_VIM = matrix(0,N,dimX1) #matrix for covariance of the number of times an obs appears in a tree and the VIM of a variable in that tree, row gives obs and columns give VIMs of each variable
covar_IJ_matrix_VIM2 = matrix(0,N,dimX1) #matrix for covariance of the number of times an obs appears in a tree and the VIM of a variable (measured as incidence of variable in that tree), row gives obs and columns give VIMs of each variable
IJ_VIM = rep(0,dimX1) #vector for variance of VIM (package computed) of each covariate across ntree trees
IJ_VIM2 = rep(0,dimX1) #vector for variance of VIM (incidence of variable in a tree) of each covariate across ntree trees

#Split Node: USING MY METHOD FOR SPLIT IMPROVEMENT AS NODE HOMOGENEITY (FOLLOWING SEGAL 1992)

VIM_SI_test = matrix(0,dimX1,ntree)       #matrix for variable importance measure created through the split improvement criteria 
VIM_SI_F_test = matrix(0,dimX1,ntree)     #matrix for variable importance measure created through the split improvement criteria tested using Hotelling T^2 converted to F statistic


VIM_SI_train = matrix(0,dimX1,ntree)       #matrix for variable importance measure created through the split improvement criteria 
VIM_SI_F_train = matrix(0,dimX1,ntree)     #matrix for variable importance measure created through the split improvement criteria tested using Hotelling T^2 converted to F statistic


VIM_Ydiff_test = array(c(0,0), dim= c(dimX1,dimY1,ntree)) #array for variable importance measure created through the differences in response
VIM_Ydiff_F_test = array(c(0,0), dim= c(dimX1,dimY1,ntree)) #array for variable importance measure created through the differences in response tested using Hotelling T^2 converted to F statistic

VIM_Ydiff_train = array(c(0,0), dim= c(dimX1,dimY1,ntree)) #array for variable importance measure created through the differences in response
VIM_Ydiff_F_train = array(c(0,0), dim= c(dimX1,dimY1,ntree)) #array for variable importance measure created through the differences in response tested using Hotelling T^2 converted to F statistic

#For loop to build ntree trees, compute prediction on oob sample for each tree
for (num_tree in (1:ntree))
{
oob_X1 = oob_X1_bootstrap[,,num_tree][which(rowSums(oob_X1_bootstrap[,,num_tree]) > 0),]

predict_oob_matrix = matrix(0,nobs_oob[num_tree],dimY1)
oob_Y1 = matrix(0,nobs_oob[num_tree],dimY1)

oob_Y1 = oob_Y1_bootstrap[,,num_tree][which(rowSums(oob_X1_bootstrap[,,num_tree]) > 0),]

#Inverse of Covariance Matrix
Inv_Cov_array1[,,num_tree] = solve(cov(trainY1_bootstrap[,,num_tree]))

#Building Single Tree
tree = build_single_tree(trainX1_bootstrap[,,num_tree], trainY1_bootstrap[,,num_tree], m_feature=10, min_leaf=10, Inv_Cov_Y = Inv_Cov_array1[,,num_tree], Command=2)

#prediction of single tree on out-of-bag samples
predict_oob_matrix = single_tree_prediction(tree, X_test=oob_X1_bootstrap[,,num_tree][which(rowSums(oob_X1_bootstrap[,,num_tree]) > 0),], Variable_number=dimY1)


#Variable Importance Measure of Each Tree, package computed variable importance
VIM_matrix1[,num_tree] = t(variable_importance_measure(tree,dimX1))


###PROPOSED VARIABLE IMPORTANCE MEASURES: USING DEFINITIONS ON SPLIT IMPROVEMENT
split_improvement_overall_train = rep(0,length(tree))   #vector for split improvement

response_difference_train = matrix(0,length(tree),dimY1) #matrix for output response difference
Hotelling_T2_train = rep(0,length(tree))        #vector for Hotelling's T2 statistic for train sample

for (node in (1:length(tree)))
{

if (length(tree[[node]]) == 5)
{

feature_no = tree[[node]][[3]]
node_threshold = tree[[node]][[4]]

nleft_train = sum((tree[[node]][[1]]>0)*1) #No. of train samples in the left node
nright_train = sum((tree[[node]][[2]]>0)*1) #No. of train samples in the right node
ntot_train = nleft_train + nright_train

#Means at parent and children nodes on train set
mean_parentnode_train = rep(0,dimY1)
mean_leftnode_train = rep(0,dimY1)
mean_rightnode_train = rep(0,dimY1)


#SSE of train sample at parent and children nodes on train set
SSE_parentnode_overall_train = rep(0,nrow(trainY1_bootstrap[,,num_tree]))
SSE_leftnode_overall_train = rep(0,nleft_train)
SSE_rightnode_overall_train = rep(0,nright_train)


for (j in (1:dimY1))
{ 
  mean_parentnode_train[j] = mean(trainY1_bootstrap[,j,num_tree]) #Mean of train sample at parent node

if (nleft_train>0)
 {
  mean_leftnode_train[j] = mean(trainY1_bootstrap[tree[[node]][[1]],j,num_tree])
}

if (nright_train>0)
{
  mean_rightnode_train[j] = mean(trainY1_bootstrap[tree[[node]][[2]],j,num_tree])
}
  
}

var_covar_parentnode_train = cov(trainY1_bootstrap[,,num_tree])

if (nleft_train>1)
{
var_covar_leftnode_train = cov(trainY1_bootstrap[tree[[node]][[1]],,num_tree])
}

if (nright_train>1)
{
var_covar_rightnode_train = cov(trainY1_bootstrap[tree[[node]][[2]],,num_tree])
}


for (num_train in (1:nrow(trainY1_bootstrap[,,num_tree])))
{
SSE_parentnode_overall_train[num_train] = (trainY1_bootstrap[num_train,,num_tree] - mean_parentnode_train)%*% ginv(var_covar_parentnode_train)%*% as.matrix(trainY1_bootstrap[num_train,,num_tree] - mean_parentnode_train)
}

if (nleft_train >0)
{
for (num_train in (1:nleft_train))
{
SSE_leftnode_overall_train[num_train]=(trainY1_bootstrap[tree[[node]][[1]][[num_train]],,num_tree] - mean_leftnode_train)%*% ginv(var_covar_parentnode_train)%*% as.matrix(trainY1_bootstrap[tree[[node]][[1]][[num_train]],,num_tree] - mean_leftnode_train)
}
}


if (nright_train >0)
{
for (num_train in (1:nright_train))
{
SSE_rightnode_overall_train[num_train]=(trainY1_bootstrap[tree[[node]][[2]][[num_train]],,num_tree] - mean_rightnode_train)%*% ginv(var_covar_parentnode_train)%*% as.matrix(trainY1_bootstrap[tree[[node]][[2]][[num_train]],,num_tree] - mean_rightnode_train)
}
}


split_improvement_overall_train[node] = sum(SSE_parentnode_overall_train) - sum(SSE_leftnode_overall_train) - sum(SSE_rightnode_overall_train)

response_difference_train[node,] = abs(mean_leftnode_train - mean_rightnode_train)

#Computing Hotelling's T^2 statistic

Hotelling_T2_train[node] = ((nleft_train*nright_train)/(nleft_train+nright_train))* (t(mean_leftnode_train - mean_rightnode_train)%*% ginv(var_covar_parentnode_train)%*% (mean_leftnode_train - mean_rightnode_train))


#Critical F
criticalF = qf(0.05,dimY1-1,samplsize -dimY1)


if (Hotelling_T2_train[node] > criticalF)
{
VIM_SI_F_train[feature_no,num_tree]= VIM_SI_F_train[feature_no,num_tree] + split_improvement_overall_train[node]  #Cumulating for the significant splits across nodes of a tree for a feature

for (j in (1:dimY1))
{
VIM_Ydiff_F_train[feature_no,j,num_tree] = VIM_Ydiff_F_train[feature_no,j,num_tree] + response_difference_train[node,j] #Cumulating for all the significant response differences across nodes of a tree for a feature
}
}

VIM_SI_train[feature_no,num_tree]= VIM_SI_train[feature_no,num_tree] + split_improvement_overall_train[node]  #Cumulating for all the split improvemenets across nodes of a tree for a feature

for (j in (1:dimY1))
{
VIM_Ydiff_train[feature_no,j,num_tree] = VIM_Ydiff_train[feature_no,j,num_tree] + response_difference_train[node,j] #Cumulating for all the response differences across nodes of a tree for a feature
}

} #Ending if parent node loop
} #Ending Node for loop

#Variance-Covariance matrix at parent and children nodes on test set 
var_covar_parentnode_test = matrix(0,dimY1,dimY1)
var_covar_parentnode_test = cov(oob_Y1-predict_oob_matrix)
ntest = nrow(oob_X1)

###RECURSIVE FUNCTION FOR TEST SET SI VIM###

TestVarImp1 = function(oob_X1,oob_Y1,tree,current.node){

        # If leaf return all zeros
        if(length(tree[[current.node]])==1 | is.null(tree[[current.node]])){ return( rep(0,dimX1) ) } else{

        # Which side do nodes go
        whichLeft = which(oob_X1[,tree[[current.node]]$Feature_number] <= tree[[current.node]]$Threshold_value) ## gives out error: Error in test_rightnode[num_test, ] : incorrect number of dimensions

        # Exception for all data goes down one side

        if( length(whichLeft) == nrow(oob_X1) ){
                return( TestVarImp1(oob_X1[whichLeft,,drop=FALSE],oob_Y1[whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][1]) ) 
        }
        if( length(whichLeft) == 0 ){
                return( TestVarImp1(oob_X1[-whichLeft,,drop=FALSE],oob_Y1[-whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][2]) ) 
        }


        #Count of nleft and nright
        nleft_test = length(whichLeft)
        test_leftnode = oob_Y1[whichLeft,,drop=FALSE]
        test_rightnode = oob_Y1[-whichLeft,,drop=FALSE]
        nright_test = nrow(as.matrix(test_rightnode))

        # Measure variable importance   
        #Means at parent and children nodes on test set
         mean_parentnode_test = rep(0,dimY1)
         mean_leftnode_test = rep(0,dimY1)
         mean_rightnode_test = rep(0,dimY1)


       #SSE of testing sample at parent node and children nodes on test set
	  SSE_parentnode_overall_test = rep(0,nrow(oob_Y1))
	  SSE_leftnode_overall_test = rep(0,length(whichLeft))
	  SSE_rightnode_overall_test = rep(0,nright_test)

       #Initializing variable importance vector
        VarImpVector = rep(0,dimX1)

       for (j in (1:dimY1))
        { 
          mean_parentnode_test[j] = mean(oob_Y1[,j]) #Mean of test sample at parent node
       
        if (nleft_test >0)
        {
          mean_leftnode_test[j] = mean(oob_Y1[whichLeft,j]) #Mean of test samples at left node based on threshold value of the node splitting variable
        }
        if (nright_test >0)
        {
          mean_rightnode_test[j] = mean(oob_Y1[-whichLeft,j]) #Mean of test samples at right node based on threshold value of node splitting variable
        }

        }

      
      for (num_test in (1:nrow(oob_Y1)))
      {
        SSE_parentnode_overall_test[num_test] = (oob_Y1[num_test,] - mean_parentnode_test)%*% ginv(var_covar_parentnode_test)%*% as.matrix(oob_Y1[num_test,] - mean_parentnode_test)
      }

       if (nleft_test >0)
        {
      for (num_test in (1:nleft_test))
        {
         SSE_leftnode_overall_test[num_test]= (test_leftnode[num_test,] - mean_leftnode_test)%*% ginv(var_covar_parentnode_test)%*% as.matrix(test_leftnode[num_test,] - mean_leftnode_test)
        }
        }

        if (nright_test >0)
        {
        for (num_test in (1:nright_test))
        {
       SSE_rightnode_overall_test[num_test]= (test_rightnode[num_test,] - mean_rightnode_test)%*% ginv(var_covar_parentnode_test)%*% as.matrix(test_rightnode[num_test,] - mean_rightnode_test)
        }
        }

       split_improvement_overall_test = sum(SSE_parentnode_overall_test) - sum(SSE_leftnode_overall_test) - sum(SSE_rightnode_overall_test)

         
        # Send whichLeft data to TestVarImp 
        leftVarImp = TestVarImp1(oob_X1[whichLeft,,drop=FALSE],test_leftnode,tree, tree[[current.node]][[5]][1])

        # send -whichLeft data to TestVarImp
        rightVarImp = TestVarImp1(oob_X1[-whichLeft,,drop=FALSE],test_rightnode,tree, tree[[current.node]][[5]][2])

        # Collate variable importances
        VarImpVector = leftVarImp + rightVarImp

        # Add this node's importance
       VarImpVector[ tree[[current.node]]$Feature_number ] = VarImpVector[ tree[[current.node]]$Feature_number ] + split_improvement_overall_test
      
}
        # return

        return( VarImpVector ) 

}

VIM_SI_test[,num_tree]  = TestVarImp1(oob_X1,oob_Y1,tree,1)

###RECURSIVE FUNCTION FOR TEST SET SI WITH F-TEST VIM##
TestVarImp2 = function(oob_X1,oob_Y1,tree,current.node){

        # If leaf return all zeros
        if(length(tree[[current.node]])==1 | is.null(tree[[current.node]])){ return( rep(0,dimX1) ) } else{

        # Which side do nodes go
        whichLeft = which(oob_X1[,tree[[current.node]]$Feature_number] <= tree[[current.node]]$Threshold_value) ## gives out error: Error in test_rightnode[num_test, ] : incorrect number of dimensions

        # Exception for all data goes down one side

        if( length(whichLeft) == nrow(oob_Y1) ){
                return( TestVarImp2(oob_X1[whichLeft,,drop=FALSE],oob_Y1[whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][1]) ) 
        }
        if( length(whichLeft) == 0 ){
                return( TestVarImp2(oob_X1[-whichLeft,,drop=FALSE],oob_Y1[-whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][2]) ) 
        }


        #Count of nleft and nright
        nleft_test = length(whichLeft)
        test_leftnode = oob_Y1[whichLeft,,drop=FALSE]
        test_rightnode = oob_Y1[-whichLeft,,drop=FALSE]
        nright_test = nrow(as.matrix(test_rightnode))

        # Measure variable importance   
        #Means at parent and children nodes on test set
         mean_parentnode_test = rep(0,dimY1)
         mean_leftnode_test = rep(0,dimY1)
         mean_rightnode_test = rep(0,dimY1)


       #SSE of testing sample at parent node and children nodes on test set
	  SSE_parentnode_overall_test = rep(0,nrow(oob_Y1))
	  SSE_leftnode_overall_test = rep(0,length(whichLeft))
	  SSE_rightnode_overall_test = rep(0,nright_test)

       #Initializing variable importance vector
        VarImpVector = rep(0,dimX1)

       for (j in (1:dimY1))
        { 
          mean_parentnode_test[j] = mean(oob_Y1[,j]) #Mean of test sample at parent node
       
        if (nleft_test >0)
        {
          mean_leftnode_test[j] = mean(oob_Y1[whichLeft,j]) #Mean of test samples at left node based on threshold value of the node splitting variable
        }
        if (nright_test >0)
        {
          mean_rightnode_test[j] = mean(oob_Y1[-whichLeft,j]) #Mean of test samples at right node based on threshold value of node splitting variable
        }

        }

      
      for (num_test in (1:nrow(oob_Y1)))
      {
        SSE_parentnode_overall_test[num_test] = (oob_Y1[num_test,] - mean_parentnode_test)%*% ginv(var_covar_parentnode_test)%*% as.matrix(oob_Y1[num_test,] - mean_parentnode_test)
      }

       if (nleft_test >0)
        {
      for (num_test in (1:nleft_test))
        {
         SSE_leftnode_overall_test[num_test]= (test_leftnode[num_test,] - mean_leftnode_test)%*% ginv(var_covar_parentnode_test)%*% as.matrix(test_leftnode[num_test,] - mean_leftnode_test)
        }
        }

        if (nright_test >0)
        {
        for (num_test in (1:nright_test))
        {
       SSE_rightnode_overall_test[num_test]= (test_rightnode[num_test,] - mean_rightnode_test)%*% ginv(var_covar_parentnode_test)%*% as.matrix(test_rightnode[num_test,] - mean_rightnode_test)
        }
        }

              
 #Computing Hotelling's T^2 statistic

if (min(nleft_test,nright_test) >= 3)
{

Hotelling_T2 = ((nleft_test*nright_test)/(nleft_test+nright_test))* (t(mean_leftnode_test - mean_rightnode_test)%*% ginv(var_covar_parentnode_test)%*% (mean_leftnode_test - mean_rightnode_test))

 #Computing F critical

criticalF = qf(0.05,dimY1-1,ntest-dimY1)} else { 
Hotelling_T2 = 0
criticalF = 0}

if (Hotelling_T2 > criticalF) { 
split_improvement_overall_test = sum(SSE_parentnode_overall_test) - sum(SSE_leftnode_overall_test) - sum(SSE_rightnode_overall_test)} else {

split_improvement_overall_test = 0}
  
        # Send whichLeft data to TestVarImp 
        leftVarImp = TestVarImp2(oob_X1[whichLeft,,drop=FALSE],test_leftnode,tree, tree[[current.node]][[5]][1])

        # send -whichLeft data to TestVarImp
        rightVarImp = TestVarImp2(oob_X1[-whichLeft,,drop=FALSE],test_rightnode,tree, tree[[current.node]][[5]][2])

        # Collate variable importances
        VarImpVector = leftVarImp + rightVarImp

        # Add this node's importance
       VarImpVector[ tree[[current.node]]$Feature_number ] = VarImpVector[ tree[[current.node]]$Feature_number ] + split_improvement_overall_test
      
}
        # return

        return( VarImpVector ) 

}

VIM_SI_F_test[,num_tree]  = TestVarImp2(oob_X1,oob_Y1,tree,1)

#RECURSIVE FUNCTION FOR TEST SET OUTCOME DIFFERENCE VIM

TestVarImp3 = function(oob_X1,oob_Y1,tree,current.node){

        # If leaf return all zeros
        if(length(tree[[current.node]])==1 | is.null(tree[[current.node]])){ return( matrix(0,dimX1,dimY1) ) } else{

        # Which side do nodes go
        whichLeft = which(oob_X1[,tree[[current.node]]$Feature_number] <= tree[[current.node]]$Threshold_value) ## gives out error: Error in test_rightnode[num_test, ] : incorrect number of dimensions

        # Exception for all data goes down one side

        if( length(whichLeft) == nrow(oob_Y1) ){
                return( TestVarImp3(oob_X1[whichLeft,,drop=FALSE],oob_Y1[whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][1]) ) 
        }
        if( length(whichLeft) == 0 ){
                return( TestVarImp3(oob_X1[-whichLeft,,drop=FALSE],oob_Y1[-whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][2]) ) 
        }


        #Count of nleft and nright
        nleft_test = length(whichLeft)
        test_leftnode = oob_Y1[whichLeft,,drop=FALSE]
        test_rightnode = oob_Y1[-whichLeft,,drop=FALSE]
        nright_test = nrow(as.matrix(test_rightnode))

        # Measure variable importance   
        #Means at parent and children nodes on test set
         mean_parentnode_test = rep(0,dimY1)
         mean_leftnode_test = rep(0,dimY1)
         mean_rightnode_test = rep(0,dimY1)
         response_difference = rep(0,dimY1)
         

       #Initializing variable importance vector
        VarImpVector = matrix(0,dimX1,dimY1)

       for (j in (1:dimY1))
        { 
          mean_parentnode_test[j] = mean(oob_Y1[,j]) #Mean of test sample at parent node
       
        if (nleft_test >0)
        {
          mean_leftnode_test[j] = mean(oob_Y1[whichLeft,j]) #Mean of test samples at left node based on threshold value of the node splitting variable
        }
        if (nright_test >0)
        {
          mean_rightnode_test[j] = mean(oob_Y1[-whichLeft,j]) #Mean of test samples at right node based on threshold value of node splitting variable
        }

        }

      response_difference = abs(mean_leftnode_test - mean_rightnode_test)

         
        # Send whichLeft data to TestVarImp 
        leftVarImp = TestVarImp3(oob_X1[whichLeft,,drop=FALSE],test_leftnode,tree, tree[[current.node]][[5]][1])

        # send -whichLeft data to TestVarImp
        rightVarImp = TestVarImp3(oob_X1[-whichLeft,,drop=FALSE],test_rightnode,tree, tree[[current.node]][[5]][2])

        # Collate variable importances
        VarImpVector = leftVarImp + rightVarImp

        # Add this node's importance
for (j in (1:dimY1))
{
       VarImpVector[ tree[[current.node]]$Feature_number, j] = VarImpVector[ tree[[current.node]]$Feature_number,j] + response_difference[j]
}
      
}
        # return

        return( VarImpVector ) 

}

VIM_Ydiff_test[,,num_tree]  = TestVarImp3(oob_X1,oob_Y1,tree,1)


##RECURSIVE FUNCTION FOR TEST SET OUTCOME DIFFERENCE WITH F-TEST VIM####

TestVarImp4 = function(oob_X1,oob_Y1,tree,current.node){

        # If leaf return all zeros
        if(length(tree[[current.node]])==1 | is.null(tree[[current.node]])){ return( matrix(0,dimX1,dimY1) ) } else{

        # Which side do nodes go
        whichLeft = which(oob_X1[,tree[[current.node]]$Feature_number] <= tree[[current.node]]$Threshold_value) ## gives out error: Error in test_rightnode[num_test, ] : incorrect number of dimensions

        # Exception for all data goes down one side

        if( length(whichLeft) == nrow(oob_Y1) ){
                return( TestVarImp4(oob_X1[whichLeft,,drop=FALSE],oob_Y1[whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][1]) ) 
        }
        if( length(whichLeft) == 0 ){
                return( TestVarImp4(oob_X1[-whichLeft,,drop=FALSE],oob_Y1[-whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][2]) ) 
        }


        #Count of nleft and nright
        nleft_test = length(whichLeft)
        test_leftnode = oob_Y1[whichLeft,,drop=FALSE]
        test_rightnode = oob_Y1[-whichLeft,,drop=FALSE]
        nright_test = nrow(as.matrix(test_rightnode))

        # Measure variable importance   
        #Means at parent and children nodes on test set
         mean_parentnode_test = rep(0,dimY1)
         mean_leftnode_test = rep(0,dimY1)
         mean_rightnode_test = rep(0,dimY1)
         response_difference = rep(0,dimY1)
         

       #Initializing variable importance vector
        VarImpVector = matrix(0,dimX1,dimY1)

       for (j in (1:dimY1))
        { 
          mean_parentnode_test[j] = mean(oob_Y1[,j]) #Mean of test sample at parent node
       
        if (nleft_test >0)
        {
          mean_leftnode_test[j] = mean(oob_Y1[whichLeft,j]) #Mean of test samples at left node based on threshold value of the node splitting variable
        }
        if (nright_test >0)
        {
          mean_rightnode_test[j] = mean(oob_Y1[-whichLeft,j]) #Mean of test samples at right node based on threshold value of node splitting variable
        }

        }

              
 #Computing Hotelling's T^2 statistic

if (min(nleft_test,nright_test) >= 3)
{

Hotelling_T2 = ((nleft_test*nright_test)/(nleft_test+nright_test))* (t(mean_leftnode_test - mean_rightnode_test)%*% ginv(var_covar_parentnode_test)%*% (mean_leftnode_test - mean_rightnode_test))

 #Computing F critical

criticalF = qf(0.05,dimY1-1,ntest-dimY1)} else { 
Hotelling_T2 = 0
criticalF = 0}

if (Hotelling_T2 > criticalF) { 
response_difference = abs(mean_leftnode_test - mean_rightnode_test)} else {

response_difference = rep(0,dimY1)}

#print(c(current.node,response_difference,nleft_test,nright_test,mean_leftnode_test,mean_rightnode_test))

         
        # Send whichLeft data to TestVarImp 
        leftVarImp = TestVarImp4(oob_X1[whichLeft,,drop=FALSE],test_leftnode,tree, tree[[current.node]][[5]][1])

        # send -whichLeft data to TestVarImp
        rightVarImp = TestVarImp4(oob_X1[-whichLeft,,drop=FALSE],test_rightnode,tree, tree[[current.node]][[5]][2])

        # Collate variable importances
        VarImpVector = leftVarImp + rightVarImp

        # Add this node's importance
for (j in (1:dimY1))
{
       VarImpVector[ tree[[current.node]]$Feature_number, j] = VarImpVector[ tree[[current.node]]$Feature_number,j] + response_difference[j]
}
      
}
        # return

        return( VarImpVector ) 

}

VIM_Ydiff_F_test[,,num_tree]  = TestVarImp4(oob_X1,oob_Y1,tree,1)
} #Ending num_tree for ntree loop


#For loop to compute the IJ variance for variable importance (as computed from package)

for (l in (1:dimX1))
{
for (k in (1:N))
{
covar_IJ_matrix_VIM[k,l] = (cov(cnt_index[k,],VIM_matrix1[l,]))^2
}
IJ_VIM[l] = sum(covar_IJ_matrix_VIM[,l])
avg_VIM1[l] = mean(VIM_matrix1[l,])
}


###AVERAGING OUT VARIABLE IMPORTANCE ACROSS ntree TREES##

for (i in (1:dimX1))
{
for (j in (1:ntree))
{
if (VIM_matrix1[i,j]>0) #Creating indicator variable if ith variable is used in jth tree

{ VIM2_matrix[i,j] = 1} else {VIM2_matrix[i,j] = 0}
}
avg_VIM2[i] = mean(VIM2_matrix[i,])
avg_VIM_SI_train[i] = mean(VIM_SI_train[i,])
avg_VIM_SI_test[i] = mean(VIM_SI_test[i,])
avg_VIM_SI_F_train[i] = mean(VIM_SI_F_train[i,])
avg_VIM_SI_F_test[i] = mean(VIM_SI_F_test[i,])
}

for (i in (1:dimX1))  
{

for (j in (1:dimY1))
{
avg_VIM_Ydiff_train[i,j] = mean(VIM_Ydiff_train[i,j,])
avg_VIM_Ydiff_test[i,j] = mean(VIM_Ydiff_test[i,j,])
avg_VIM_Ydiff_F_train[i,j] = mean(VIM_Ydiff_F_train[i,j,])
avg_VIM_Ydiff_F_test[i,j] = mean(VIM_Ydiff_F_test[i,j,])
}

}


#Exporting outputs

myfile1 <- file.path(filepath, paste0(num_forest, "_avg_VIM_count3K",  ".csv"))
myfile2 <- file.path(filepath, paste0(num_forest, "_avg_VIM_ind3K",  ".csv"))
myfile3 <- file.path(filepath, paste0(num_forest, "_avg_VIM_SI_oob3K",  ".csv"))
myfile4 <- file.path(filepath, paste0(num_forest, "_avg_VIM_SIF_oob3K",  ".csv"))
myfile5 <- file.path(filepath, paste0(num_forest, "_avg_VIM_Ydiff_train3K",  ".csv"))
myfile6 <- file.path(filepath, paste0(num_forest, "_avg_VIM_Ydiff_test_oob3K",  ".csv"))
myfile7 <- file.path(filepath, paste0(num_forest, "_avg_VIM_Ydiff_F_train3K",  ".csv"))
myfile8 <- file.path(filepath, paste0(num_forest, "_avg_VIM_Ydiff_F_test_oob3K",  ".csv"))

write.csv(avg_VIM1, file = myfile1)
write.csv(avg_VIM2, file = myfile2)
write.csv(cbind(avg_VIM_SI_train,avg_VIM_SI_test), file = myfile3)
write.csv(cbind(avg_VIM_SI_F_train,avg_VIM_SI_F_test), file = myfile4)
write.csv(avg_VIM_Ydiff_train, file = myfile5)
write.csv(avg_VIM_Ydiff_test, file = myfile6)
write.csv(avg_VIM_Ydiff_F_train, file = myfile7)
write.csv(avg_VIM_Ydiff_F_test, file = myfile8)


} # loop ending for num_forest





















