#install.packages("rpart")
#install.packages('randomForest')
#install.packages('MultivariateRandomForest')
#install.packages('MulvariateRandomForestVarImp')
#install.packages("boot")
#install.packages("hydroGOF")
#install.packages("rlang", dependencies = TRUE, INSTALL_opts = '--no-lock')
#install.packages("rlang")
#install.packages("devtools")
#devtools::install_github("Megatvini/VIM")


require(rpart)
require(randomForest)
require(MulvariateRandomForestVarImp)
require(MultivariateRandomForest)
require(boot)
require(plyr)
require(hydroGOF)
require(MASS)

#Importing datasets

data(EbirdData)


trainY1 = EbirdData$trainY[,2:6]
trainX1 = EbirdData$trainX[,2:86]
testY1 = EbirdData$testY[,2:6]
testX1 = EbirdData$testX[,2:86]


N=nrow(trainX1)
ntest = nrow(testX1)

#USER-DEFINED INPUTS
#specify where you want results to store
filepath = './'

#reduce number of trees to 2 and iterations to 1 for code testing
ntree = 2 #user defined number of trees, for the paper Sikdar et al. this was set at 2000
niter= 1 #user defined number of iterations, for the paper Sikdar et al. this was set at 20


##BOOTSTRAPPING TRAINING AND TESTING SAMPLES
samplsize =  round(0.035*nrow(trainX1))
dimY1 = ncol(trainY1)
dimX1 = ncol(trainX1)+ 1
results1 = matrix(0,samplsize,ntree)  #Initializing matrix of indices of observations used in each tree

for (i in 1:ntree)
{
## A bootstrap sample of indices of observations for each tree
results1[,i]= t(sample(1:N,size = samplsize, replace=TRUE))
}

for (iter in (1:niter))
{
##BOOTSTRAPPING TRAINING AND TESTING SAMPLES

cnt_index = matrix(0,N,ntree)         #Initializing matrix of counts of indices of observations used in each tree, i.e., cnt_index[1,1] is the number of times 1st obs was used in tree 1
oob_index = matrix(0,N,ntree)
nobs_oob = rep(0,ntree) #vector for the number of oob samples in ntree trees

###Building Trees
###Initializing arrays and matrices
Inv_Cov_array1 = array(c(0,0), dim= c(dimY1,dimY1,ntree))
predict_test_array1 = array(c(0,0), dim= c(ntest,dimY1,ntree))
sumsqrderr_test = matrix(0,ntree,dimY1)
sumsqrderr_oob = matrix(0,ntree,dimY1)
MAE_mat = rep(0,dimY1)
RMSE_mat = rep(0,dimY1)
SE_mat = rep(0,dimY1)



###################################################################################################
#######PART 2 OF THE EXERCISE: VIM  USING SPLIT IMPROVEMENT BY COMPARING PARENT AND CHILDREN NODES#######
###################################################################################################


#Split Node: USING MY METHOD FOR SPLIT IMPROVEMENT AS NODE HOMOGENEITY (FOLLOWING SEGAL 1992)

VIM_Ydiff = array(c(0,0), dim= c(dimX1,dimY1,ntree)) #array for variable importance measure created through the differences in response

#For loop to build ntree trees, compute prediction and MSE on test sample and oob sample for each tree
for (i in (1:ntree))
{

trainX1_bootstrap = matrix(0, samplsize,dimX1)
trainY1_bootstrap = matrix(0, samplsize,dimY1)
oob_X1_bootstrap = matrix(0, N,dimX1)
oob_Y1_bootstrap = matrix(0, N, dimY1)

#ADDING NOISE COLUMN IN THE COVARIATES MATRIX
noise1 = runif(nrow(trainX1), 0, 1)
trainX11= cbind(trainX1,noise1)

for (j in (1:samplsize))
{ 
##Creating bootstrap samples based on indices of observations used in each tree
trainX1_bootstrap[j,] = as.numeric(trainX11[results1[j,i],])
trainY1_bootstrap[j,] = as.numeric(trainY1[results1[j,i],])
}


for (k in 1:N)
{
cnt_index[k,i] = sum((k == results1[,i])*1)

if (cnt_index[k,i] == 0)
{oob_index[k,i] = k
 oob_X1_bootstrap[k,] = as.numeric(trainX11[k,])
 oob_Y1_bootstrap[k,] = as.numeric(trainY1[k,])} else {
 oob_index[k,i] = 0
 oob_X1_bootstrap[k,] = 0
 oob_Y1_bootstrap[k,] = 0
}}

nobs_oob[i] = length(which(rowSums(oob_X1_bootstrap[,]) > 0))

oob_X1 = oob_X1_bootstrap[which(rowSums(oob_X1_bootstrap) > 0),]

predict_oob_matrix = matrix(0,nobs_oob[i],dimY1)
oob_Y1 = matrix(0,nobs_oob[i],dimY1)

oob_Y1 = oob_Y1_bootstrap[which(rowSums(oob_X1_bootstrap) > 0),]


#Inverse of Covariance Matrix
Inv_Cov_array1[,,i] = solve(cov(trainY1_bootstrap))

#Building Single Tree
tree = build_single_tree(trainX1_bootstrap, trainY1_bootstrap, m_feature=10, min_leaf=10, Inv_Cov_Y = Inv_Cov_array1[,,i], Command=2)

#Prediction of Single Tree on test sample
noise2 = runif(nrow(testX1), 0, 1)
testX11=cbind(testX1,noise2)


predict_test_array1[,,i] = single_tree_prediction(tree, X_test=testX11, Variable_number=dimY1)

#prediction of single tree on out-of-bag samples
predict_oob_matrix = single_tree_prediction(tree, X_test=oob_X1, Variable_number=dimY1)

for (m in (1:dimY1))
{
sumsqrderr_test[i,m] = sum((testY1[,m]-predict_test_array1[,m,i])^2) 
sumsqrderr_oob[i,m] = sum((oob_Y1[,m]-predict_oob_matrix[,m])^2)
}

TestVarImp = function(oob_X1,oob_Y1,tree,current.node){

        # If leaf return all zeros
        if(length(tree[[current.node]])==1 | is.null(tree[[current.node]])){ return( matrix(0,dimX1,dimY1) ) } else{

        # Which side do nodes go
        whichLeft = which(oob_X1[,tree[[current.node]]$Feature_number] <= tree[[current.node]]$Threshold_value) ## gives out error: Error in test_rightnode[num_test, ] : incorrect number of dimensions

        # Exception for all data goes down one side

        if( length(whichLeft) == nrow(oob_Y1) ){
                return( TestVarImp(oob_X1[whichLeft,,drop=FALSE],oob_Y1[whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][1]) ) 
        }
        if( length(whichLeft) == 0 ){
                return( TestVarImp(oob_X1[-whichLeft,,drop=FALSE],oob_Y1[-whichLeft,,drop=FALSE],tree, tree[[current.node]][[5]][2]) ) 
        }


        #Count of nleft and nright
        nleft_test = length(whichLeft)
        test_leftnode = oob_Y1[whichLeft,,drop=FALSE]
        test_rightnode = oob_Y1[-whichLeft,,drop=FALSE]
        nright_test = nrow(as.matrix(test_rightnode))

        # Measure variable importance   
        #Means at parent and children nodes on test set
         mean_parentnode_test = rep(0,dimY1)
         mean_leftnode_test = rep(0,dimY1)
         mean_rightnode_test = rep(0,dimY1)
         response_difference = rep(0,dimY1)
         

       #Initializing variable importance vector
        VarImpVector = matrix(0,dimX1,dimY1)

       for (j in (1:dimY1))
        { 
          mean_parentnode_test[j] = mean(oob_Y1[,j]) #Mean of test sample at parent node
       
        if (nleft_test >0)
        {
          mean_leftnode_test[j] = mean(oob_Y1[whichLeft,j]) #Mean of test samples at left node based on threshold value of the node splitting variable
        }
        if (nright_test >0)
        {
          mean_rightnode_test[j] = mean(oob_Y1[-whichLeft,j]) #Mean of test samples at right node based on threshold value of node splitting variable
        }

        }

      response_difference = abs(mean_leftnode_test - mean_rightnode_test)

         
        # Send whichLeft data to TestVarImp 
        leftVarImp = TestVarImp(oob_X1[whichLeft,,drop=FALSE],test_leftnode,tree, tree[[current.node]][[5]][1])

        # send -whichLeft data to TestVarImp
        rightVarImp = TestVarImp(oob_X1[-whichLeft,,drop=FALSE],test_rightnode,tree, tree[[current.node]][[5]][2])

        # Collate variable importances
        VarImpVector = leftVarImp + rightVarImp

        # Add this node's importance
for (j in (1:dimY1))
{
       VarImpVector[ tree[[current.node]]$Feature_number, j] = VarImpVector[ tree[[current.node]]$Feature_number,j] + response_difference[j]
}
      
}
        # return

        return( VarImpVector ) 

}

VIM_Ydiff[,,i]  = TestVarImp(oob_X1,oob_Y1,tree,1)
print(i)

} #Ending i for ntree loop



###AVERAGING OUT VARIABLE IMPORTANCE AND NOISE DIFFERENCE COMPUTED AS ABOVE ACROSS ntree TREES##


avg_VIM_Ydiff = matrix(0,dimX1,dimY1) #Mean variable importance computed using the response difference method
var_VIM_Ydiff = matrix(0,dimX1,dimY1) #Mean variable importance computed using the response difference method

for (i in (1:dimX1))  
{
for (j in (1:dimY1))
{
avg_VIM_Ydiff[i,j] = mean(VIM_Ydiff[i,j,])
var_VIM_Ydiff[i,j] = var(VIM_Ydiff[i,j,])
}
}


#For loop to compute the IJ variance for variable importance as computed above 

covar_IJ_matrix_VIM_Ydiff = array(c(0,0), dim= c(N,dimX1,dimY1)) #array for covariance of the number of times an obs appears in a tree and the VIM_Ydiff of a variable (measured as incidence of variable in that tree), row gives obs and columns give VIMs of each variable

IJ_VIM_Ydiff = matrix(0,dimX1,dimY1) #matrix for variance of VIM_Ydiff of each covariate across ntree trees
IJ_VIM_Ydiff_corr = matrix(0,dimX1,dimY1) #matrix for variance of VIM_Ydiff of each covariate across ntree trees

for (l in (1:dimX1))
{
for (j in (1:dimY1))
{
for (k in (1:N))
{
covar_IJ_matrix_VIM_Ydiff[k,l,j] = (cov(cnt_index[k,],VIM_Ydiff[l,j,]))^2
}

IJ_VIM_Ydiff[l,j] = sum(covar_IJ_matrix_VIM_Ydiff[,l,j])
IJ_VIM_Ydiff_corr[l,j]= IJ_VIM_Ydiff[l,j]-(samplsize/ntree)*var_VIM_Ydiff[l,j]

}}

###Averaging out predictions across ntree trees

avg_predict1 = matrix(0,ntest,dimY1)
for (i in (1:ntest))
{
for (j in (1:dimY1))
{
avg_predict1[i,j] = mean(predict_test_array1[i,j,])
}
}

##COMPUTING MAE AND RMSE and OOB error
for (j in (1:dimY1))
{
MAE_mat[j] = mean(abs(avg_predict1[,j]-testY1[,j]))
RMSE_mat[j] = mean((avg_predict1[,j]-testY1[,j])^2)
SE_mat[j] = sd((avg_predict1[,j]-testY1[,j])^2)/sqrt(nrow(testY1))
}

retained_cov_indx1 = (1:dimX1)[avg_VIM_Ydiff[,1]> avg_VIM_Ydiff[dimX1,1]]
retained_cov_indx2 = (1:dimX1)[avg_VIM_Ydiff[,2]> avg_VIM_Ydiff[dimX1,2]]
retained_cov_indx3 = (1:dimX1)[avg_VIM_Ydiff[,3]> avg_VIM_Ydiff[dimX1,3]]
retained_cov_indx4 = (1:dimX1)[avg_VIM_Ydiff[,4]> avg_VIM_Ydiff[dimX1,4]]
retained_cov_indx5 = (1:dimX1)[avg_VIM_Ydiff[,5]> avg_VIM_Ydiff[dimX1,5]]
retained_cov_indx = unique(c(retained_cov_indx1,retained_cov_indx2,retained_cov_indx3,retained_cov_indx4,retained_cov_indx5))

dimp = length(retained_cov_indx)

#UPDATING TRAINING AND TESTING SETS

trainX1 = trainX1[,retained_cov_indx[1:dimp]]
testX1 = testX11[,retained_cov_indx[1:dimp]]

#UPDATING VIM LIST FOR THIS ITERATION
avg_VIM_Ydiff = rbind(avg_VIM_Ydiff[retained_cov_indx[1:dimp],],avg_VIM_Ydiff[dimX1,])
IJ_VIM_Ydiff = rbind(IJ_VIM_Ydiff[retained_cov_indx[1:dimp],],IJ_VIM_Ydiff[dimX1,])
IJ_VIM_Ydiff_corr = rbind(IJ_VIM_Ydiff_corr[retained_cov_indx[1:dimp],],IJ_VIM_Ydiff_corr[dimX1,])

#UPDATING ALL VALUES
dimX1 = ncol(trainX1)+ 1
dimX1_iterations = dimX1-1

#WRITING OUTPUT TO FILE PATH LOCATION
myfile <- file.path(filepath, paste0(iter, "_avg_predict",  ".csv"))
myfile2 <- file.path(filepath, paste0(iter, "covariates",  ".csv"))
myfile3 <- file.path(filepath, paste0(iter, "avgVIM",  ".csv"))
myfile4 <- file.path(filepath, paste0(iter, "MAE_mat",  ".csv"))
myfile5 <- file.path(filepath, paste0(iter, "MSE_mat",  ".csv"))
myfile6 <- file.path(filepath, paste0(iter, "dimX1_iterations",  ".csv"))
myfile7 <- file.path(filepath, paste0(iter, "IJ_VIM_corr",  ".csv"))
myfile8 <- file.path(filepath, paste0(iter, "SE_mat",  ".csv"))


write.csv(avg_predict1, file = myfile)
write.csv(head(trainX1), file = myfile2)
write.csv(avg_VIM_Ydiff, file = myfile3)
write.csv(MAE_mat, file = myfile4)
write.csv(RMSE_mat, file = myfile5)
write.csv(dimX1_iterations, file = myfile6)
write.csv(IJ_VIM_Ydiff_corr, file = myfile7)
write.csv(SE_mat, file = myfile8)

print(iter)

} #END OF ITERATION LOOP


