#install.packages("rpart")
#install.packages('randomForest')
#install.packages('MultivariateRandomForest')
#install.packages('MulvariateRandomForestVarImp')
#install.packages("boot")
#install.packages("hydroGOF")
#install.packages("rlang", dependencies = TRUE, INSTALL_opts = '--no-lock')
#install.packages("rlang")
#install.packages("devtools")
#devtools::install_github("Megatvini/VIM")


require(rpart)
require(randomForest)
require(MulvariateRandomForestVarImp)
require(MultivariateRandomForest)
require(boot)
require(plyr)
require(hydroGOF)
require(MASS)

#Importing datasets

data(EbirdData)


trainY1 = EbirdData$trainY[,2:6]
trainX1 = EbirdData$trainX[,2:86]
testY1 = EbirdData$testY[,2:6]
testX1 = EbirdData$testX[,2:86]

N=nrow(trainX1)
ntest = nrow(testX1)

#USER-DEFINED INPUTS
#specify where you want results to store
filepath = './'

#reduce number of trees to 2 and iterations to 1 for code testing
ntree = 2 #user defined number of trees, for the paper Sikdar et al. this was set at 2000
niter=1 #user defined number of iterations, for the paper Sikdar et al. this was set at 20

##BOOTSTRAPPING TRAINING AND TESTING SAMPLES
samplsize =  round(0.035*nrow(trainX1))
dimY1 = ncol(trainY1)
dimX1 = ncol(trainX1)+ 1
results1 = matrix(0,samplsize,ntree)  #Initializing matrix of indices of observations used in each tree

for (i in 1:ntree)
{
## A bootstrap sample of indices of observations for each tree
results1[,i]= t(sample(1:N,size = samplsize, replace=TRUE))

}

for (iter in (1:niter))
{

##BOOTSTRAPPING TRAINING AND TESTING SAMPLES

cnt_index = matrix(0,N,ntree)         #Initializing matrix of counts of indices of observations used in each tree, i.e., cnt_index[1,1] is the number of times 1st obs was used in tree 1
oob_index = matrix(0,N,ntree)
nobs_oob = rep(0,ntree) #vector for the number of oob samples in ntree trees

###Building Trees
###Initializing arrays and matrices
Inv_Cov_array1 = array(c(0,0), dim= c(dimY1,dimY1,ntree))
predict_test_array1 = array(c(0,0), dim= c(ntest,dimY1,ntree))
VIM_matrix1 = matrix(0,dimX1,ntree)
VIM2_matrix = matrix(0,dimX1,ntree) #indicator matrix where (i,j)th value takes 1 if ith covariate is in jth tree, else 0
avg_VIM2 = rep(0,dimX1) #mean variable importance across ntree trees
var_VIM2 = rep(0,dimX1) #mean variable importance across ntree trees
sumsqrderr_test = matrix(0,ntree,dimY1)
sumsqrderr_oob = matrix(0,ntree,dimY1)
MAE_mat = rep(0,dimY1)
RMSE_mat = rep(0,dimY1)
SE_mat = rep(0,dimY1)
covar_IJ_matrix_VIM2 = matrix(0,N,dimX1) #matrix for covariance of the number of times an obs appears in a tree and the VIM of a variable (measured as incidence of variable in that tree), row gives obs and columns give VIMs of each variable
IJ_VIM2 = rep(0,dimX1) #vector for variance of VIM (incidence of variable in a tree) of each covariate across ntree trees
IJ_VIM2_corr = rep(0,dimX1) #vector for variance of VIM (incidence of variable in a tree) of each covariate across ntree trees after bias correction


for (i in 1:ntree)
{


trainX1_bootstrap = matrix(0, samplsize,dimX1)
trainY1_bootstrap = matrix(0, samplsize,dimY1)
oob_X1_bootstrap = matrix(0, N,dimX1)
oob_Y1_bootstrap = matrix(0, N, dimY1)

#ADDING NOISE COLUMN IN THE COVARIATES MATRIX
noise1 = runif(nrow(trainX1), 0, 1)
trainX11= cbind(trainX1,noise1)

for (j in (1:samplsize))
{ 
##Creating bootstrap samples based on indices of observations used in each tree
trainX1_bootstrap[j,] = as.numeric(trainX11[results1[j,i],])
trainY1_bootstrap[j,] = as.numeric(trainY1[results1[j,i],])
}


for (k in 1:N)
{
cnt_index[k,i] = sum((k == results1[,i])*1)

if (cnt_index[k,i] == 0)
{oob_index[k,i] = k
 oob_X1_bootstrap[k,] = as.numeric(trainX11[k,])
 oob_Y1_bootstrap[k,] = as.numeric(trainY1[k,])} else {
 oob_index[k,i] = 0
 oob_X1_bootstrap[k,] = 0
 oob_Y1_bootstrap[k,] = 0
}}

nobs_oob[i] = length(which(rowSums(oob_X1_bootstrap[,]) > 0))

#Inverse of Covariance Matrix
Inv_Cov_array1[,,i] = solve(cov(trainY1_bootstrap))

#Building Single Tree
model2 = build_single_tree(trainX1_bootstrap, trainY1_bootstrap, m_feature=min(10,dimX1), min_leaf=10, Inv_Cov_Y = Inv_Cov_array1[,,i], Command=2)

#Prediction of Single Tree on test sample
noise2 = runif(nrow(testX1), 0, 1)
testX11=cbind(testX1,noise2)
predict_test_array1[,,i] = single_tree_prediction(model2, X_test=testX11, Variable_number=dimY1)

for (m in (1:dimY1))
{
sumsqrderr_test[i,m] = sum((testY1[,m]-predict_test_array1[,m,i])^2) 
}

#Variable Importance Measure of Each Tree, package computed variable importance
VIM_matrix1[,i] = t(variable_importance_measure(model2,dimX1))
print(i)

}


###AVERAGING OUT VARIABLE IMPORTANCE ACROSS ntree TREES##

for (i in (1:dimX1))
{
for (j in (1:ntree))
{
if (VIM_matrix1[i,j]>0) #Creating indicator variable if ith variable is used in jth tree

{ VIM2_matrix[i,j] = 1} else {VIM2_matrix[i,j] = 0}
}
avg_VIM2[i] = mean(VIM2_matrix[i,])
var_VIM2[i] = var(VIM2_matrix[i,])
}

#For loop to compute the IJ variance for variable importance as computed above as indicator of whether variable is used in a tree

for (l in (1:dimX1))
{
for (k in (1:N))
{
covar_IJ_matrix_VIM2[k,l] = (cov(cnt_index[k,],VIM2_matrix[l,]))^2
}
IJ_VIM2[l] = sum(covar_IJ_matrix_VIM2[,l])
IJ_VIM2_corr[l] = IJ_VIM2[l]-(samplsize/ntree)*var_VIM2[l]
}


###Averaging out predictions across ntree trees

avg_predict1 = matrix(0,ntest,dimY1)
for (i in (1:ntest))
{
for (j in (1:dimY1))
{
avg_predict1[i,j] = mean(predict_test_array1[i,j,])
}
}

####COMPUTING MAE AND RMSE###

for (j in (1:dimY1))
{
MAE_mat[j] = mean(abs(avg_predict1[,j]- testY1[,j]))
RMSE_mat[j] = mean((avg_predict1[,j]- testY1[,j])^2)
SE_mat[j] = sd((avg_predict1[,j]-testY1[,j])^2)/sqrt(nrow(testY1))
}


retained_cov_indx = (1:dimX1)[avg_VIM2> avg_VIM2[dimX1]]
dimp = length(retained_cov_indx)

#print(avg_VIM2)
#print(dimp)
#print(retained_cov_indx)
#print(dimX1)
#print(ncol(trainX1))
if (dimp!=0) {
#UPDATING TRAINING AND TESTING SETS
trainX1 = trainX11[,retained_cov_indx[1:dimp]]
testX1 = testX11[,retained_cov_indx[1:dimp]]

#UPDATING VIM LIST  AND IJ VARIANCE AND CORRECTED VARIANCE FOR THIS ITERATION

avg_VIM2 = c(avg_VIM2[retained_cov_indx[1:dimp]],avg_VIM2[dimX1])
IJ_VIM2 = c(IJ_VIM2[retained_cov_indx[1:dimp]],IJ_VIM2[dimX1])
IJ_VIM2_corr = c(IJ_VIM2_corr[retained_cov_indx[1:dimp]],IJ_VIM2_corr[dimX1])

}

else {
#UPDATING TRAINING AND TESTING SETS
trainX1 = trainX1
testX1 = testX1

#UPDATING VIM LIST  AND IJ VARIANCE AND CORRECTED VARIANCE FOR THIS ITERATION

avg_VIM2 = avg_VIM2
IJ_VIM2 = IJ_VIM2
IJ_VIM2_corr = IJ_VIM2_corr[dimX1]
}


#UPDATING ALL VALUES

dimX1 = ncol(trainX1)+ 1
dimX1_iterations = dimX1-1

#CHANGE FILEPATH TO LOCATION WHERE YOU WANT RESULTS TO STORE

myfile <- file.path(filepath, paste0(iter, "_avg_predict",  ".csv"))
myfile2 <- file.path(filepath, paste0(iter, "covariates",  ".csv"))
myfile3 <- file.path(filepath, paste0(iter, "avgVIM",  ".csv"))
myfile4 <- file.path(filepath, paste0(iter, "MAE_mat",  ".csv"))
myfile5 <- file.path(filepath, paste0(iter, "MSE_mat",  ".csv"))
myfile6 <- file.path(filepath, paste0(iter, "dimX1_iterations",  ".csv"))
myfile7 <- file.path(filepath, paste0(iter, "IJ_VIM_corr",  ".csv"))
myfile8 <- file.path(filepath, paste0(iter, "SE_mat",  ".csv"))



write.csv(avg_predict1, file = myfile)
write.csv(head(trainX1), file = myfile2)
write.csv(avg_VIM2, file = myfile3)
write.csv(MAE_mat, file = myfile4)
write.csv(RMSE_mat, file = myfile5)
write.csv(dimX1_iterations, file = myfile6)
write.csv(IJ_VIM_corr, file = myfile7)
write.csv(SE_mat, file = myfile8)


print(iter)

} #END OF ITERATION LOOP




